import datetime
import functools
import operator
import sys

import kanboard
import neaty.log as LOG
import saturnin.context

import pykbrd


class Condition:

    def __init__(self, key, value, desc='(some condition)'):
        self.key = key
        self.value = value
        self._op = operator.eq  # that's all we support for now
        self.desc = desc

    def _error_itemkey(self, item, key):
        def guesstype(item):
            url = item.get('url', 'P://H/p/(unknown item type)/n')
            return url.split('/')[-2]
        raise pykbrd.InternalError(
            "invalid condition application: %s data has no key %r (%s)"
            % (guesstype(item), key, self)
        )

    def match(self, item):
        if self.key not in item:
            self._error_itemkey(item, self.key)
        return self._op(item[self.key], self.value)


class ConditionSet:

    def __init__(self, conditions):
        self.conditions = conditions

    def filter(self, items):
        return [i for i in items if self.match(i)]

    def match(self, item):
        if not self.conditions:
            return True
        for cond in self.conditions:
            if not cond.match(item):
                return False
        return True


class Param:

    def __init__(self, kw, factory=None,
                 fmt_desc=None, mandatory=False,
                 default=None):
        self.kw = kw
        self._factory = factory
        self.fmt_desc = fmt_desc or ''
        self.mandatory = mandatory
        self._default = default

    @property
    def desc(self):
        return (
            "%s[kw=%s]"
            % (self.__class__.__name__, self.kw)
        )

    def chop_from(self, args):
        if not args:
            return {}, args
        if not args[0] == self.kw:
            return {}, args
        if not len(args) > 1:
            raise pykbrd.UsageError("missing value for: %s" % self.kw)
        key, value, rest = args[0], args[1], args[2:]
        return {key: self.init(value)}, rest

    def emit_default(self, args):
        if self._default is None:
            return {}
        if callable(self._default):
            return {self.kw: self._default()}
        return {self.kw: self._default}

    def init(self, value):
        if self._factory is None:
            return value
        try:
            return self._factory(value)
        except Exception:
            raise pykbrd.UsageError("bad format: %s must be %s"
                                    % (self.kw, self.fmt_desc))

    def verify_mandatory(self, parsed):
        """
        Check if *parsed* params contain our param, if it's mandatory
        """
        if not self.mandatory:
            return None
        if self.kw in parsed:
            return None
        raise pykbrd.UsageError("missing mandatory parameter: %s"
                                % self.desc)

    def parse(self, args):
        out = self.emit_default(args)
        chopped, rest = self.chop_from(args)
        out.update(chopped)
        self.verify_mandatory(chopped)
        return out, rest


class PositionalParam(Param):

    def __init__(self, kw, factory=None,
                 fmt_desc=None, mandatory=False,
                 default=None):
        self.kw = kw
        self._factory = factory
        self.fmt_desc = fmt_desc or ''
        self.mandatory = mandatory
        self._default = default

    def chop_from(self, args):
        if not args:
            return {}, args
        value, rest = args[0], args[1:]
        return {self.kw: self.init(value)}, rest


class BoolParam(Param):

    mandatory = False

    def __init__(self, kw, inv_kw=None, default=False):
        self.kw = kw
        self._default = default
        self._kws = {
            kw: not default,
            inv_kw: default,
        }

    def chop_from(self, args):
        if not args:
            return {}, args
        if not args[0] in self._kws:
            return {}, args
        value, rest = args[0], args[1:]
        return {self.kw: self.init(value)}, rest

    def init(self, value):
        return self._kws.get(value)


class ParamSet:

    def __init__(self, params, unknown=None, terminator=None):
        self._params = params
        self._unknown = unknown or (lambda a: True)
        self._terminator = terminator

    def emit_default(self, args):
        out = {}
        for param in self._params:
            out.update(param.emit_default(args))
        return out

    def _check_unknown(self, args):
        if not self._unknown:
            return
        if not args:
            return
        if not self._unknown(args[0]):
            return
        raise pykbrd.UsageError("unknown argument: %s" % args[0])

    def _can_chop(self, args):
        if not args:
            return False
        if not self._terminator:
            return True
        return not self._terminator(args[0])

    def _chop_one(self, args):
        rest = args[:]
        for param in self._params:
            new, rest = param.chop_from(rest)
            if new:
                return new, rest
        return {}, rest

    def chop_from(self, args):
        rest = args[:]
        out = {}
        while self._can_chop(rest):
            chop, rest = self._chop_one(rest)
            if not chop:
                break
            out.update(chop)
        self._check_unknown(rest)
        return out, rest

    def parse(self, args):
        out = self.emit_default(args)
        chopped, rest = self.chop_from(args)
        out.update(chopped)
        self.verify_mandatory(chopped)
        return out, rest

    def verify_mandatory(self, parsed):
        """
        Check if *parsed* params contain our param, if it's mandatory
        """
        for p in self._params:
            p.verify_mandatory(parsed)


class AltParamSet(ParamSet):

    def __init__(self, schemes, mandatory=False):
        self._schemes = schemes
        self._picks = {}
        self.mandatory = mandatory

    def _pick_scheme(self, args):
        case = repr(args)
        if case not in self._picks:
            for scheme in self._schemes:
                LOG.debug('trying scheme: %r' % scheme)
                try:
                    out, rest = scheme.parse(args[:])
                except pykbrd.UsageError as e:
                    LOG.debug('..that failed: %s' % e)
                    continue
                self._picks[case] = scheme
        if case not in self._picks:
            raise pykbrd.UsageError("no usage pattern matched: %r" % args)
        return self._picks[case]

    @property
    def desc(self):
        return '|'.join([s.desc for s in self._schemes])

    def chop_from(self, args):
        return self._pick_scheme(args).chop_from(args)

    def emit_default(self, args):
        return {}
        return self._pick_scheme(args).emit_default(args)

    def parse(self, args):
        return self._pick_scheme(args).parse(args)

    def verify_mandatory(self, parsed):
        if not self.mandatory:
            return
        if bool(parsed):
            return
        raise pykbrd.UsageError(
            "missing one of mandatory parameters: %s"
            % self.desc
        )


class App:

    @classmethod
    def from_argv(cls, argv):
        """
        Parse *argv* and return tuple containing App instance and list
        of remaining arguments.

        Obtain options by parsing arguments *argv* (list of strings,
        typically sys.argv) using APP_SCHEME.  Return tuple containing
        two elements: the new App instance, and a list instance containing
        rest of the arguments; these should be used to instantiate action
        in the sub-command.
        """
        LOG.debug('argv[1:]=%r' % argv[1:])
        try:
            options, action_args = APP_SCHEME.parse(argv[1:])
        except pykbrd.UsageError as e:
            LOG.warn(str(e))
            sys.exit(2)
        LOG.debug('action_args=%r' % action_args)
        if options['--dry']:
            cls = DryApp
        return cls(options=options), action_args

    def __init__(self, options=None):
        self.options = options or {}
        self.config = Config.from_saturnin()
        self._kc = None
        LOG.think("App.options=%r" % self.options)

    @property
    def kc(self):
        """
        Kanboard Client instance.
        """
        if self._kc is None:
            self._kc = self._connect()
        return self._kc

    def _connect(self):
        """
        Actually create the instance of Client.

        Basically kanboard.Client, but wrapped in a custom class
        that helps intercept queries for purposes of debugging and
        optimization.
        """
        def redact(token):
            return token[:4] + '[REDACTED]' + token[-4:]
        LOG.debugv('self.config.baseurl', self.config.baseurl)
        LOG.debugv('self.config.username', self.config.username)
        LOG.debugv('self.config.token', redact(self.config.token))
        return Client(
            self.config.baseurl,
            self.config.username,
            self.config.token,
        )

    def _execute(self, actionfn, actionparams):
        LOG.debugv('actionfn', actionfn)
        LOG.debugv('actionparams', actionparams)
        result = self.kc.execute(actionfn, **actionparams)
        LOG.debugv('result', result)
        return result

    def run_action(self, action):
        try:
            return action.run(self)
        except pykbrd.InternalError as e:
            LOG.warn(str(e))
            sys.exit(4)
        except pykbrd.ExternalError as e:
            LOG.warn(str(e))
            sys.exit(3)


class DryApp(App):
    """
    pykbrd.app.App sub-class that will not connect anywhere, just print
    what would be done.

    This should be used in place of App when user passed --dry argument.
    """

    def _dry_log(self, msg):
        LOG.warn('dry: ' + msg)

    def _connect(self):
        baseurl = self.config.baseurl
        username = self.config.username
        token = '(redacted)'
        self._dry_log("kc = Client(%r, %r, %r)" % (baseurl, username, token))
        return True

    def _execute(self, actionfn, actionparams):
        self._connect()
        self._dry_log("kc.execute(%r, **%r)" % (actionfn, actionparams))


class _BaseAction:

    @classmethod
    def new(cls, scheme, args,
            ki=None, base_params=None, formatter_cls=None):
        params = dict(base_params or {})
        try:
            cli_params, rest = scheme.parse(args)
        except pykbrd.UsageError as e:
            LOG.warn(str(e))
            sys.exit(2)
        if rest:
            raise pykbrd.UsageError("extra arguments: %r" % rest)
        params |= cli_params
        LOG.think("%s.params=%r" % (cls.__name__, params))
        return cls(
            ki=ki,
            params=params,
            formatter_cls=formatter_cls,
        )

    def __init__(self, params, ki=None, formatter_cls=None):
        self.ki = ki
        self.params = params
        self.formatter_cls = formatter_cls
        self._loader = None
        self._editor = None

    def _seen_any(self, *keys):
        """
        True if any of provided 'keys' were specified in params.
        """
        for k in keys:
            if k in self.params:
                return True
        return False

    def make_loader(self, kc):
        """
        Kanboard object loader using Client and current params.
        """
        if self._loader is None:
            self._loader = Loader(kc, self.params)
        return self._loader

    def make_editor(self, kc):
        """
        Kanboard object editor using Client and current params.
        """
        if self._editor is None:
            self._editor = Editor(kc, self.params)
        return self._editor

    def run(self, app):
        raise NotImplementedError


class CreationAction(_BaseAction):
    """
    Generic action representing creation of a Kanboard object.

    Object type is specified by providing name of the Kanboard API method.
    """

    def run(self, app):
        response = app._execute(self.ki.create_method.name, self.params)
        if response is False:
            raise pykbrd.ExternalError(
                "object creation failed: method=%r, params=%r"
                % (self.ki.create_method.name, self.params)
            )
        return ActionResult(
            data=response,
            lines=[str(response)],
        )


class TaskMoveAction(_BaseAction):
    """
    Moving of a Kanboard task.
    """

    targets = [
        'column_id',
        'position',
        'swimlane_id',
    ]

    def prep_params(self, app):
        task = app.kc.getTask(task_id=self.params['task_id'])
        if not self._seen_any(*self.targets):
            raise pykbrd.UsageError("need to specify either of: %s"
                                    % ', '.join(self.targets))
        identity = {
            'task_id': task['id'],
            'project_id': task['project_id'],
        }
        state = {
            k: task[k]
            for k in self.targets
        }
        LOG.debugv('identity', identity)
        LOG.debugv('state', state)
        return identity | state | self.params

    def run(self, app):
        response = app._execute(
            'moveTaskPosition',
            self.prep_params(app),
        )
        if response is False:
            raise pykbrd.ExternalError(
                "task move failed: method=%r, params=%r"
                % (self.ki.move_method.name, self.params)
            )
        return ActionResult(
            data=response,
            lines=[],
        )


class ListingAction(_BaseAction):
    """
    Generic action representing listing of a set of Kanboard objects.

    Object type is specified by providing name of the Kanboard API method.
    """

    def _get_data(self, loader):
        data = loader.load_list(self.ki)
        cset = loader.load_conditions()
        LOG.debugv('len(data)', len(data))
        fdata = data.filter_by_condition(cset)
        LOG.debugv('len(fdata)', len(fdata))
        return fdata

    def run(self, app):
        loader = self.make_loader(app.kc)
        formatter = self.formatter_cls(loader=loader)
        data = self._get_data(loader)
        return ActionResult(
            data=data,
            lines=formatter.list_lines(data),
        )


class DeletionAction(_BaseAction):
    """
    Generic action representing deletion of a Kanboard object.

    Object type is specified by providing name of the Kanboard API method.
    """

    def run(self, app):
        editor = self.make_editor(app.kc)
        response = editor.delete_item(self.ki)
        return ActionResult(
            data=response,
            lines=[],
        )


class ShowingAction(_BaseAction):
    """
    Generic action representing showing of a Kanboard object.

    Object type is specified by providing name of the Kanboard API method.
    """

    def run(self, app):
        loader = self.make_loader(app.kc)
        item = loader.load_item(self.ki)
        formatter = self.formatter_cls(loader=loader)
        return ActionResult(
            data=item,
            lines=formatter.page_lines(item),
        )


class ActionResult:

    def __init__(self, data, lines):
        self.data = data
        self.lines = lines


class Config:

    @classmethod
    def from_saturnin(cls):
        api_cfg_data = (
            saturnin.context
            .get_config_repo(filename='api.ini')
            .data('api')
        )
        baseurl = api_cfg_data['baseurl'][0]
        username = api_cfg_data['username'][0]
        token = api_cfg_data['token'][0]
        return cls({
            'baseurl': baseurl,
            'username': username,
            'token': token,
        })

    def __init__(self, opts={}):
        self._opts = opts
        for k in opts:
            setattr(self, k, opts[k])


class Client(kanboard.Client):

    def __init__(self, *args, **kwargs):
        self._born = datetime.datetime.now()
        self._last = self._born
        super().__init__(*args, **kwargs)

    def texec(self, mname, **kwargs):
        ex_start = datetime.datetime.now()
        since = ex_start - self._last
        age = ex_start - self._born
        self._last = ex_start
        result = self.execute(mname, **kwargs)
        duration = datetime.datetime.now() - ex_start
        LOG.warn('%02.3f | +%02.3f | %s(%s) | %02.3f' % (
            age.total_seconds(),
            since.total_seconds(),
            mname, kwargs,
            duration.total_seconds(),
        ))
        return result


class ItemList(list):

    def __init__(self, s, ki=None):
        super().__init__(s)
        self.ki = ki

    def filter(self, fn):
        """
        Clone self, keeping only items where *fn*, is true
        """
        return self.__class__([i for i in self if fn(i)])

    def filter_by_key(self, key, value):
        """
        Clone self, keeping only items where *key* key equals *value*
        """
        return self.filter(
            fn=lambda c: c[key] == value,
        )

    def filter_by_condition(self, condition):
        """
        Clone self, keeping only items where *condition* matches
        """
        return self.filter(
            fn=lambda i: condition.match(i)
        )

    def pick(self, fn, desc):
        """
        Select exactly one item where *fn*, described by *desc* is true

        If none or more than one are selected, raise InternalError.
        """
        selected = [i for i in self if fn(i)]
        if len(selected) == 0:
            raise pykbrd.InternalError(
                "object not found: %s where %s" % (self.ki.noun_s, desc)
            )
        if len(selected) > 1:
            raise pykbrd.InternalError(
                "too many objects found: %s where %s" % (self.ki.noun_s, desc)
            )
        LOG.debugv('selected', selected)
        return selected[0]

    def pick_by_key(self, key, value):
        """
        Select exactly one item where *key* key equals *value*.

        If none or more than one are selected, raise InternalError.
        """
        return self.pick(
            fn=lambda c: c[key] == value,
            desc="%s %s is '%s'" % (self.ki.noun_s, key, value),
        )

    def pick_by_condition(self, condition):
        """
        Clone self, keeping only items where *condition* matches

        If none or more than one are selected, raise InternalError.
        """
        return self.pick(
            fn=lambda i: condition.match(i),
            desc=condition.desc,
        )


class Loader:

    def __init__(self, kc, params):
        self.kc = kc
        self.params = params

    @functools.lru_cache
    def _fetch(self, method, **kwargs):
        return self.kc.texec(method, **kwargs)

    def load_item(self, ki, **params):
        method = ki.fetch_method
        method_params = {
            k: self.params[k] for k in self.params
            if k in method.requires
        } | params
        return self._fetch(
            method.name,
            **method_params,
        )

    def load_list(self, ki):
        method = ki.list_method
        method_params = {
            k: self.params[k] for k in self.params
            if k in method.requires
        }
        return ItemList(self._fetch(
            method.name,
            **method_params,
        ), ki)

    def _load_condition(self, raw_key, raw_value):
        desc = '%s %s' % (raw_key, raw_value)
        if raw_key == ':swimlane':
            swimlane = self.swimlanes().pick_by_key(
                key='name',
                value=raw_value,
            )
            return Condition('swimlane_id', swimlane['id'], desc)
        if raw_key == ':column':
            column = self.columns().pick_by_key(
                key='title',
                value=raw_value,
            )
            return Condition('column_id', column['id'], desc)
        raise pykbrd.InternalError("unsupported filter: %s" % raw_key)

    def load_conditions(self):
        conditions = []
        for k, v in self.params.items():
            if k.startswith(':'):
                conditions.append(self._load_condition(k, v))
        return ConditionSet(conditions)

    def swimlanes(self):
        return self.load_list(pykbrd.kbrd.SWIMLANE_INFO)

    def columns(self):
        return self.load_list(pykbrd.kbrd.COLUMN_INFO)

    def projects(self):
        return self.load_list(pykbrd.kbrd.PROJECT_INFO)

    def tasks(self):
        return self.load_list(pykbrd.kbrd.TASK_INFO)

    def task_by_id(self, id_):
        return self.load_item(pykbrd.kbrd.TASK_INFO, task_id=id_)

    def project_by_id(self, id_):
        return self.load_item(pykbrd.kbrd.PROJECT_INFO, project_id=id_)

    def swimlane_by_id(self, id_):
        return self.load_item(pykbrd.kbrd.SWIMLANE_INFO, swimlane_id=id_)

    def column_by_id(self, id_):
        return self.load_item(pykbrd.kbrd.COLUMN_INFO, column_id=id_)


class Editor(Loader):

    def delete_item(self, ki):
        method = ki.delete_method
        method_name = method.name
        method_params = {
            k: self.params[k] for k in self.params
            if k in method.requires
        }
        LOG.warnv('method_name', method_name)
        LOG.warnv('method_params', method_params)
        return self.kc.texec(method_name, **method_params)


def connect():
    c = Config.from_saturnin()
    return Client(c.baseurl, c.username, c.token)


APP_SCHEME = ParamSet(
    params=[
        BoolParam(kw='--dry', default=False),
        Param(kw='--foo', factory=str, default='Foo'),
    ],
    unknown=lambda a: a.startswith('--'),
    terminator=lambda a: not a.startswith('--')
)
