import datetime
import functools

import neaty.log as LOG
import pykbrd.table
import pykbrd


class _BaseFormatter:

    def __init__(self, loader, opts=None):
        self.loader = loader
        self.opts = opts or {}
        self.rowdelim = ' '

    def lines(self, data):
        return []

    def body(self, data):
        return '\n'.join(self.lines(data)) + '\n'

    def format_table(self, fields, data, delim=None, sort=None):
        def sortt(R):
            return tuple(R[key] for key in sort)
        sort = sort or []
        delim = delim or self.rowdelim
        header = [fields[f] or f for f in fields]
        rows = [header]
        for rd in sorted(data, key=sortt):
            row = [rd[f] for f in fields]
            rows.append(row)
        return pykbrd.table.Table.from_data(rows).fmt()

    def decorate(self, item):
        """
        Insert human-readable variants of essential related fields
        """
        raise NotImplementedError("implement me in a sub-class!")


class TaskFormatter(_BaseFormatter):

    def decorate(self, task):
        """
        Insert human-readable variants of essential related fields
        """
        swimlane = self.loader.swimlane_by_id(task['swimlane_id'])
        project = self.loader.project_by_id(task['project_id'])
        column = self.loader.column_by_id(task['column_id'])
        return task | {
            'column_position': column['position'],
            'column_title': column['title'],
            'project_name': project['name'],
            'swimlane_name': swimlane['name'],
            'swimlane_position': swimlane['position'],
        }

    def page_lines(self, data):
        def lrender(F):
            out.append(F % page_data)
        page_data = self.decorate(data)
        out = []
        lrender("task id: %(id)s")
        lrender("    swimlane: %(swimlane_name)s; column: %(column_title)s")
        lrender("    title: %(title)s")
        if page_data['description']:
            out.append("    description:")
            out.extend("        " + ln for ln in
                       page_data['description'].splitlines())
        lrender("    url: %(url)s")
        return out

    def list_lines(self, data):
        tbl_data = [self.decorate(t) for t in data]
        fields = {
            'id': 'ID',
            'project_name': 'PROJECT',
            'swimlane_name': 'SWIMLANE',
            'column_title': 'COLUMN',
            'title': 'TITLE',
        }
        sort = [
            'project_name',
            'swimlane_position',
            'column_position',
            'position',
            'id',
        ]
        return self.format_table(fields, tbl_data, sort=sort)


class SwimlaneFormatter(_BaseFormatter):

    def decorate(self, swimlane):
        """
        Insert human-readable variants of essential related fields
        """
        project = self.loader.project_by_id(swimlane['project_id'])
        return swimlane | {
            'project_name': project['name'],
            'is_active_bool': bool(swimlane['is_active']),
        }

    def list_lines(self, data):
        data = data or []
        tbl_data = [self.decorate(s) for s in data]
        fields = {
            'project_name': 'PROJECT',
            'id': 'ID',
            'name': 'NAME',
            'description': 'DESCRIPTION',
        }
        sort = [
            'project_name',
            'position',
            'id',
        ]
        return self.format_table(fields, tbl_data, sort=sort)

    def page_lines(self, data):
        def lrender(F):
            out.append(F % page_data)
        page_data = self.decorate(data)
        out = []
        lrender("swimlane id: %(id)s")
        lrender("       name: %(name)s")
        lrender("    project: %(project_name)s")
        lrender("   position: %(position)s")
        lrender("     active: %(is_active_bool)s")
        return out


class ColumnFormatter(_BaseFormatter):

    def decorate(self, column):
        """
        Insert human-readable variants of essential related fields
        """
        project = self.loader.project_by_id(column['project_id'])
        return column | {
            'project_name': project['name'],
        }

    def list_lines(self, data):
        data = data or []
        tbl_data = [self.decorate(c) for c in data]
        fields = {
            'project_name': 'PROJECT',
            'id': 'ID',
            'title': 'TITLE',
        }
        sort = [
            'project_name',
            'position',
            'id',
        ]
        return self.format_table(fields, tbl_data, sort=sort)

    def page_lines(self, data):
        def lrender(F):
            out.append(F % page_data)
        page_data = self.decorate(data)
        out = []
        lrender(" column id: %(id)s")
        lrender("     title: %(title)s")
        lrender("   project: %(project_name)s")
        lrender("  position: %(position)s")
        lrender("task limit: %(task_limit)s")
        return out


class ProjectFormatter(_BaseFormatter):

    def decorate(self, project):
        """
        Insert human-readable variants of essential related fields
        """
        lmdate = datetime.datetime.fromtimestamp(project['last_modified'])
        return project | {
            'last_modified_h': lmdate.isoformat(),
        }

    def list_lines(self, data):
        data = data or []
        tbl_data = [self.decorate(p) for p in data]
        fields = {
            'id': 'ID',
            'name': 'NAME',
            'last_modified_h': 'LAST_MODIFIED',
            'description': 'DESCRIPTION',
        }
        sort = [
            'id',
        ]
        return self.format_table(fields, tbl_data, sort=sort)
